==> gm-1.6.3_Linux.tar.gz: gm-1.6.3_Linux/libgm/gm_page_alloc.c <==
/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/* author: glenn@myri.com */

/* This modules implements a platform-independent page allocation
   interface, with automatic initialization and finalization for use
   in the kernel.

   In some cases, this file uses platform-specific features to
   implement page allocation efficiently, but by default this module
   allocates blocks of pages and keeps a page free list to recycle
   freed pages, but the module never frees the blocks of pages until
   all pages have been freed.  */

#include "gm_call_trace.h"
#include "gm_debug.h"
#include "gm_internal.h"
#include "gm_malloc_debug.h"

struct gm_free_page
{
  struct gm_free_page *next;
};

struct gm_page_allocation_record
{
  struct gm_page_allocation_record *next;
  void *buf;
};

static struct gm_free_page *_gm_first_free_page;
static struct gm_page_allocation_record *_gm_first_page_allocation_record;
static unsigned long _gm_page_allocation_cnt;

/* Allocate a page-aligned page. */

static
void *
_gm_page_alloc (void)
{
  struct gm_free_page *page;

  GM_CALLED ();
  
#if GM_KERNEL && GM_KERNEL_MALLOC_ALIGNS_PAGES
  page = (struct gm_free_page *) _gm_malloc (GM_PAGE_LEN);
#elif !GM_KERNEL && GM_HAVE_MEMALIGN
  page = (struct gm_free_page *) memalign (GM_PAGE_LEN, GM_PAGE_LEN);
#else
  if (_gm_first_free_page)
    {
      /* Return a page in the free list. */

      page = _gm_first_free_page;
      _gm_first_free_page = _gm_first_free_page->next;
    }
  else
    {
      int i;
      void *temp;
      struct gm_page_allocation_record *rec;

      /* Allocate more free pages, plus room for the allocation record. */

      if (!GM_PAGE_LEN)
	{
	  GM_NOTE (("gm_alloc_page: do not know GM_PAGE_LEN.\n"));
	  GM_RETURN_PTR (0);
	}
      temp = _gm_malloc (33 * GM_PAGE_LEN - 1);
      if (!temp)
	{
	  GM_RETURN_PTR (0);
	}
      
      /* Determine location of first new page. */

      if (((gm_size_t) temp & (GM_PAGE_LEN - 1)) == 0)
	page = temp;
      else
	page = (struct gm_free_page *)
	  ((char *) temp + GM_PAGE_LEN - ((gm_size_t) temp & (GM_PAGE_LEN - 1)));

      /* Determine where to store allocation record, utilizing the
         otherwised wasted space allocated to ensure we could align
         32 pages. */

      if ((gm_size_t) ((char *) page - (char *) temp) >= sizeof (*rec))
	rec = (struct gm_page_allocation_record *) temp;
      else
	rec = ((struct gm_page_allocation_record *)
	       ((char *) page + 32 * GM_PAGE_LEN));

      /* Record allocation */

      rec->buf = temp;
      rec->next = _gm_first_page_allocation_record;
      _gm_first_page_allocation_record = rec;

      /* Add 31 of the 32 new free pages to the free list. */

      for (i = 0; i < 31; i++)
	{
	  page->next = _gm_first_free_page;
	  _gm_first_free_page = page;
	  page = (struct gm_free_page *) ((char *) page + GM_PAGE_LEN);
	}
    }
  _gm_page_allocation_cnt++;

  /* Return the remaining page to the caller. */
#endif

  /* allocation succeeded */

  gm_assert (page);
  gm_assert (GM_PAGE_LEN);
  gm_assert (GM_POWER_OF_TWO (GM_PAGE_LEN));
  gm_assert (GM_PAGE_ALIGNED (page));

  GM_RETURN_PTR (page);
}

/* Free a page.  If all pages have been freed, free all the memory
   allocated for pages. */

static
void
_gm_page_free (void *addr)
{
  GM_CALLED ();

  gm_assert (GM_PAGE_LEN);
  gm_assert (GM_POWER_OF_TWO (GM_PAGE_LEN));
  gm_assert (GM_PAGE_ALIGNED (addr));

#if GM_KERNEL && GM_KERNEL_MALLOC_ALIGNS_PAGES
  _gm_free (addr);
#elif !GM_KERNEL && GM_HAVE_MEMALIGN
  free (addr);
#else
  {
    struct gm_free_page *page;

    gm_assert ((((gm_size_t) addr & (GM_PAGE_LEN - 1)) == 0));
    page = (struct gm_free_page *) addr;
    page->next = _gm_first_free_page;
    _gm_first_free_page = page;
    _gm_page_allocation_cnt--;
    if (!_gm_page_allocation_cnt)
      {
	while (_gm_first_page_allocation_record)
	  {
	    struct gm_page_allocation_record *next;

	    next = _gm_first_page_allocation_record->next;
	    _gm_free (_gm_first_page_allocation_record->buf);
	    _gm_first_page_allocation_record = next;
	  }
	_gm_first_free_page = 0;
	GM_PRINT (GM_PRINT_LEVEL >= 1,
		  ("Freed all memory reserved for pages.\n"));
      }
  }
#endif
  GM_RETURN_NOTHING ();
}

GM_ENTRY_POINT void *
gm_page_alloc ()
{
  void *ret;

  GM_CALLED ();
  ret = _gm_page_alloc ();
  if (!ret)
    goto abort_with_nothing;
  if (GM_MDEBUG_RECORD_PTR (ret, gm_page_free) != GM_SUCCESS)
    goto abort_with_ret;
  GM_RETURN (ret);

 abort_with_ret:
  _gm_page_free (ret);
 abort_with_nothing:
  GM_RETURN (0);
}

GM_ENTRY_POINT
void
gm_page_free (void *ptr)
{
  GM_CALLED ();
  GM_MDEBUG_REMOVE_PTR (ptr, gm_page_free);
  _gm_page_free (ptr);
  GM_RETURN_NOTHING ();
}


/*
  This file uses GM standard indentation:

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  End:
*/

==> gm-2.0.8_Linux.tar.gz: gm-2.0.8_Linux/libgm/gm_page_alloc.c <==
/******************************************************************-*-c-*-
 * Myricom GM networking software and documentation                      *
 * Copyright (c) 1996, 1997 by Myricom, Inc.                             *
 * All rights reserved.  See the file `COPYING' for copyright notice.    *
 *************************************************************************/

/**
   @file gm_page_alloc.c
   This file contains gm_page_alloc() and gm_page_free().
 
   These GM API functions allow pages to be allocated and freed.
 
   These modules implement a platform-independent page allocation
   interface, with automatic initialization and finalization for use
   in the kernel.
 
   In some cases, this file uses platform-specific features to
   implement page allocation efficiently, but by default this module
   allocates blocks of pages and keeps a page free list to recycle
   freed pages, but the module never frees the blocks of pages until
   all pages have been freed.
*/

#include "gm_call_trace.h"
#include "gm_debug.h"
#include "gm_internal.h"
#include "gm_malloc_debug.h"

/** A structure representing a free page.  This structure is placed at the
    start of each free page, and used to link these free pages in a list. */
struct gm_free_page
{
  struct gm_free_page *next;
};

/** A record of a large buffer allocation used to allocate a bunch of
    pages.  (We allocate pages in blocks of many pages, in case the OS's
    memory allocator does not give us a page-aligned allocation, and
    use only the aligned pages within this large buffer. */
struct gm_page_allocation_record
{
  struct gm_page_allocation_record *next;
  void *buf;
};

static struct gm_free_page *_gm_first_free_page;
static struct gm_page_allocation_record *_gm_first_page_allocation_record;
static unsigned long _gm_page_allocation_cnt;

/* Allocate a page-aligned page. */

static
void *
_gm_page_alloc (void)
{
  struct gm_free_page *page;

  GM_CALLED ();
  
#if GM_KERNEL && GM_KERNEL_MALLOC_ALIGNS_PAGES
  page = (struct gm_free_page *) _gm_malloc (GM_PAGE_LEN);
#elif !GM_KERNEL && GM_HAVE_MEMALIGN
  page = (struct gm_free_page *) memalign (GM_PAGE_LEN, GM_PAGE_LEN);
#else
  if (_gm_first_free_page)
    {
      /* Return a page in the free list. */

      page = _gm_first_free_page;
      _gm_first_free_page = _gm_first_free_page->next;
    }
  else
    {
      int i;
      void *temp;
      struct gm_page_allocation_record *rec;

      /* Allocate more free pages, plus room for the allocation record. */

      if (!GM_PAGE_LEN)
	{
	  GM_NOTE (("gm_alloc_page: do not know GM_PAGE_LEN.\n"));
	  GM_RETURN_PTR (0);
	}
      temp = _gm_malloc (33 * GM_PAGE_LEN - 1);
      if (!temp)
	{
	  GM_RETURN_PTR (0);
	}
      
      /* Determine location of first new page. */

      if (((gm_size_t) temp & (GM_PAGE_LEN - 1)) == 0)
	page = temp;
      else
	page = (struct gm_free_page *)
	  ((char *) temp + GM_PAGE_LEN - ((gm_size_t) temp & (GM_PAGE_LEN - 1)));

      /* Determine where to store allocation record, utilizing the
         otherwised wasted space allocated to ensure we could align
         32 pages. */

      if ((gm_size_t) ((char *) page - (char *) temp) >= sizeof (*rec))
	rec = (struct gm_page_allocation_record *) temp;
      else
	rec = ((struct gm_page_allocation_record *)
	       ((char *) page + 32 * GM_PAGE_LEN));

      /* Record allocation */

      rec->buf = temp;
      rec->next = _gm_first_page_allocation_record;
      _gm_first_page_allocation_record = rec;

      /* Add 31 of the 32 new free pages to the free list. */

      for (i = 0; i < 31; i++)
	{
	  page->next = _gm_first_free_page;
	  _gm_first_free_page = page;
	  page = (struct gm_free_page *) ((char *) page + GM_PAGE_LEN);
	}
    }
  _gm_page_allocation_cnt++;

  /* Return the remaining page to the caller. */
#endif

  /* allocation succeeded */

  gm_assert (page);
  gm_assert (GM_PAGE_LEN);
  gm_assert (GM_POWER_OF_TWO (GM_PAGE_LEN));
  gm_assert (GM_PAGE_ALIGNED (page));

  GM_RETURN_PTR (page);
}

/* Free a page.  If all pages have been freed, free all the memory
   allocated for pages. */

static
void
_gm_page_free (void *addr)
{
  GM_CALLED ();

  gm_assert (GM_PAGE_LEN);
  gm_assert (GM_POWER_OF_TWO (GM_PAGE_LEN));
  gm_assert (GM_PAGE_ALIGNED (addr));

#if GM_KERNEL && GM_KERNEL_MALLOC_ALIGNS_PAGES
  _gm_free (addr);
#elif !GM_KERNEL && GM_HAVE_MEMALIGN
  free (addr);
#else
  {
    struct gm_free_page *page;

    gm_assert ((((gm_size_t) addr & (GM_PAGE_LEN - 1)) == 0));
    page = (struct gm_free_page *) addr;
    page->next = _gm_first_free_page;
    _gm_first_free_page = page;
    _gm_page_allocation_cnt--;
    if (!_gm_page_allocation_cnt)
      {
	while (_gm_first_page_allocation_record)
	  {
	    struct gm_page_allocation_record *next;

	    next = _gm_first_page_allocation_record->next;
	    _gm_free (_gm_first_page_allocation_record->buf);
	    _gm_first_page_allocation_record = next;
	  }
	_gm_first_free_page = 0;
	GM_PRINT (GM_PRINT_LEVEL >= 1,
		  ("Freed all memory reserved for pages.\n"));
      }
  }
#endif
  GM_RETURN_NOTHING ();
}

/**
   gm_page_alloc() returns a ptr to a newly allocated page-aligned buffer
   of length GM_PAGE_LEN.
 
   @retval ptr Page-aligned buffer of length GM_PAGE_LEN.
 
   @see gm_page_free
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT void *
gm_page_alloc ()
{
  void *ret;

  GM_CALLED ();
  ret = _gm_page_alloc ();
  if (!ret)
    goto abort_with_nothing;
  if (GM_MDEBUG_RECORD_PTR (ret, gm_page_free) != GM_SUCCESS)
    goto abort_with_ret;
  GM_RETURN (ret);

 abort_with_ret:
  _gm_page_free (ret);
 abort_with_nothing:
  GM_RETURN (0);
}

/**
   gm_page_free() frees the page of memory at \b ptr previously
   allocated by gm_page_alloc().  If all pages have been freed, free
   all of the memory allocated for pages.
 
   @param ptr \b (IN) Address of the memory page to be freed.
 
   @see gm_page_alloc
 
   @author Glenn Brown
 
   @version GM_API_VERSION (as defined in gm.h)
 
*/

GM_ENTRY_POINT
void
gm_page_free (void *ptr)
{
  GM_CALLED ();
  GM_MDEBUG_REMOVE_PTR (ptr, gm_page_free);
  _gm_page_free (ptr);
  GM_RETURN_NOTHING ();
}


/*
  This file uses GM standard indentation:

  Local Variables:
  c-file-style:"gnu"
  tab-width:8
  End:
*/
